# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import tempfile

from portage.const import BASH_BINARY
from portage.tests import TestCase
from portage.util import ensure_dirs
from portage.util._dyn_libs.dyn_libs import installed_dynlibs
from portage.util.file_copy import copyfile


class InstalledDynlibsTestCase(TestCase):
    def testInstalledDynlibsRegular(self):
        """
        Return True for *.so regular files.
        """
        with tempfile.TemporaryDirectory() as directory:
            bash_copy = os.path.join(directory, "lib", "libfoo.so")
            ensure_dirs(os.path.dirname(bash_copy))
            copyfile(BASH_BINARY, bash_copy)
            self.assertTrue(installed_dynlibs(directory))

    def testInstalledDynlibsOnlySymlink(self):
        """
        If a *.so symlink is installed but does not point to a regular
        file inside the top directory, installed_dynlibs should return
        False (bug 921170).
        """
        with tempfile.TemporaryDirectory() as directory:
            symlink_path = os.path.join(directory, "lib", "libfoo.so")
            ensure_dirs(os.path.dirname(symlink_path))
            os.symlink(BASH_BINARY, symlink_path)
            self.assertFalse(installed_dynlibs(directory))

    def testInstalledDynlibsSymlink(self):
        """
        Return True for a *.so symlink pointing to a regular file inside
        the top directory.
        """
        with tempfile.TemporaryDirectory() as directory:
            bash_copy = os.path.join(directory, BASH_BINARY.lstrip(os.sep))
            ensure_dirs(os.path.dirname(bash_copy))
            copyfile(BASH_BINARY, bash_copy)
            symlink_path = os.path.join(directory, "lib", "libfoo.so")
            ensure_dirs(os.path.dirname(symlink_path))
            os.symlink(bash_copy, symlink_path)
            self.assertTrue(installed_dynlibs(directory))

    def testInstalledDynlibsAbsoluteSymlink(self):
        """
        If a *.so symlink target is outside of the top directory,
        traversal follows the corresponding file inside the top
        directory if it exists, and otherwise stops following the
        symlink.
        """
        with tempfile.TemporaryDirectory() as directory:
            bash_copy = os.path.join(directory, BASH_BINARY.lstrip(os.sep))
            ensure_dirs(os.path.dirname(bash_copy))
            copyfile(BASH_BINARY, bash_copy)
            symlink_path = os.path.join(directory, "lib", "libfoo.so")
            ensure_dirs(os.path.dirname(symlink_path))
            os.symlink(BASH_BINARY, symlink_path)
            self.assertTrue(installed_dynlibs(directory))



# Copyright 2011-2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase

from portage.checksum import checksum_str, _apply_hash_filter
from portage.exception import DigestException


class ChecksumTestCase(TestCase):
    text = b"Some test string used to check if the hash works"

    def test_md5(self):
        self.assertEqual(checksum_str(b"", "MD5"), "d41d8cd98f00b204e9800998ecf8427e")
        self.assertEqual(
            checksum_str(self.text, "MD5"), "094c3bf4732f59b39d577e9726f1e934"
        )

    def test_sha1(self):
        self.assertEqual(
            checksum_str(b"", "SHA1"), "da39a3ee5e6b4b0d3255bfef95601890afd80709"
        )
        self.assertEqual(
            checksum_str(self.text, "SHA1"), "5c572017d4e4d49e4aa03a2eda12dbb54a1e2e4f"
        )

    def test_sha256(self):
        self.assertEqual(
            checksum_str(b"", "SHA256"),
            "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
        )
        self.assertEqual(
            checksum_str(self.text, "SHA256"),
            "e3d4a1135181fe156d61455615bb6296198e8ca5b2f20ddeb85cb4cd27f62320",
        )

    def test_sha512(self):
        self.assertEqual(
            checksum_str(b"", "SHA512"),
            "cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e",
        )
        self.assertEqual(
            checksum_str(self.text, "SHA512"),
            "c8eaa902d48a2c82c2185a92f1c8bab8115c63c8d7a9966a8e8e81b07abcb9762f4707a6b27075e9d720277ba9fec072a59840d6355dd2ee64681d8f39a50856",
        )

    def test_rmd160(self):
        try:
            self.assertEqual(
                checksum_str(b"", "RMD160"), "9c1185a5c5e9fc54612808977ee8f548b2258d31"
            )
            self.assertEqual(
                checksum_str(self.text, "RMD160"),
                "fc453174f63fc011d6f64abd2c45fb6a53c8239b",
            )
        except DigestException:
            self.skipTest("RMD160 implementation not available")

    def test_whirlpool(self):
        try:
            self.assertEqual(
                checksum_str(b"", "WHIRLPOOL"),
                "19fa61d75522a4669b44e39c1d2e1726c530232130d407f89afee0964997f7a73e83be698b288febcf88e3e03c4f0757ea8964e59b63d93708b138cc42a66eb3",
            )
            self.assertEqual(
                checksum_str(self.text, "WHIRLPOOL"),
                "8f556a079b87057f19e0880eed6d833e40c916f4b133196f6842281a2517873074d399832470c11ee251696b4844a10197714a069ba3e3415c8a4eced8f91b48",
            )
        except DigestException:
            self.skipTest("WHIRLPOOL implementation not available")

    def test_blake2b(self):
        try:
            self.assertEqual(
                checksum_str(b"", "BLAKE2B"),
                "786a02f742015903c6c6fd852552d272912f4740e15847618a86e217f71f5419d25e1031afee585313896444934eb04b903a685b1448b755d56f701afe9be2ce",
            )
            self.assertEqual(
                checksum_str(self.text, "BLAKE2B"),
                "84cb3c88838c7147bc9797c6525f812adcdcb40137f9c075963e3a3ed1fe06aaeeb4d2bb5589bad286864dc1aa834cfc4d66b8d7e4d4a246d91d45ce3a6eee43",
            )
        except DigestException:
            self.skipTest("BLAKE2B implementation not available")

    def test_blake2s(self):
        try:
            self.assertEqual(
                checksum_str(b"", "BLAKE2S"),
                "69217a3079908094e11121d042354a7c1f55b6482ca1a51e1b250dfd1ed0eef9",
            )
            self.assertEqual(
                checksum_str(self.text, "BLAKE2S"),
                "823ab2429f27690450efe888b0404d092fe2ee72a9bd63d5342c251b4dbb373d",
            )
        except DigestException:
            self.skipTest("BLAKE2S implementation not available")

    def test_sha3_256(self):
        try:
            self.assertEqual(
                checksum_str(b"", "SHA3_256"),
                "a7ffc6f8bf1ed76651c14756a061d662f580ff4de43b49fa82d80a4b80f8434a",
            )
            self.assertEqual(
                checksum_str(self.text, "SHA3_256"),
                "932fc0498ebb865406f9b6606280939283aa8a148562e39fd095a5d22bdec5c6",
            )
        except DigestException:
            self.skipTest("SHA3_256 implementation not available")

    def test_sha3_512(self):
        try:
            self.assertEqual(
                checksum_str(b"", "SHA3_512"),
                "a69f73cca23a9ac5c8b567dc185a756e97c982164fe25859e0d1dcc1475c80a615b2123af1f5f94c11e3e9402c3ac558f500199d95b6d3e301758586281dcd26",
            )
            self.assertEqual(
                checksum_str(self.text, "SHA3_512"),
                "6634c004dc31822fa65c2f1e2e3bbf0cfa35085653cca1ca9ca42f8f3f13c908405e0b665918146181c9fc9a9d793fc05429d669c35a55517820dfaa071425ca",
            )
        except DigestException:
            self.skipTest("SHA3_512 implementation not available")


class ApplyHashFilterTestCase(TestCase):
    def test_apply_hash_filter(self):
        indict = {"MD5": "", "SHA1": "", "SHA256": "", "size": ""}

        self.assertEqual(
            sorted(_apply_hash_filter(indict, lambda x: True)),
            ["MD5", "SHA1", "SHA256", "size"],
        )
        self.assertEqual(
            sorted(_apply_hash_filter(indict, lambda x: x == "MD5")), ["MD5", "size"]
        )
        self.assertEqual(
            sorted(_apply_hash_filter(indict, lambda x: x != "MD5")),
            ["SHA1", "SHA256", "size"],
        )
        self.assertEqual(
            sorted(_apply_hash_filter(indict, lambda x: x == "SHA256")),
            ["SHA256", "size"],
        )
        # this should return size + one of the hashes
        self.assertEqual(len(list(_apply_hash_filter(indict, lambda x: False))), 2)



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.exception import InvalidData
from portage.localization import _
from portage.dep.soname.SonameAtom import SonameAtom

_error_empty_category = _("Multilib category empty: %s")
_error_missing_category = _("Multilib category missing: %s")
_error_duplicate_category = _("Multilib category occurs" " more than once: %s")


def parse_soname_deps(s):
    """
    Parse a REQUIRES or PROVIDES dependency string, and raise
    InvalidData if necessary.

    @param s: REQUIRES or PROVIDES string
    @type s: str
    @rtype: iter
    @return: An iterator of SonameAtom instances
    """

    categories = set()
    category = None
    previous_soname = None
    for soname in s.split():
        if soname.endswith(":"):
            if category is not None and previous_soname is None:
                raise InvalidData(_error_empty_category % category)

            category = soname[:-1]
            previous_soname = None
            if category in categories:
                raise InvalidData(_error_duplicate_category % category)
            categories.add(category)

        elif category is None:
            raise InvalidData(_error_missing_category % soname)
        else:
            previous_soname = soname
            yield SonameAtom(category, soname)

    if category is not None and previous_soname is None:
        raise InvalidData(_error_empty_category % category)



"""
Abstract base classes for rich path objects.

This module is published as a PyPI package called "pathlib-abc".

This module is also a *PRIVATE* part of the Python standard library, where
it's developed alongside pathlib. If it finds success and maturity as a PyPI
package, it could become a public part of the standard library.

Two base classes are defined here -- PurePathBase and PathBase -- that
resemble pathlib's PurePath and Path respectively.
"""

import functools
from glob import _Globber, _no_recurse_symlinks
from errno import ENOENT, ENOTDIR, EBADF, ELOOP, EINVAL
from stat import S_ISDIR, S_ISLNK, S_ISREG, S_ISSOCK, S_ISBLK, S_ISCHR, S_ISFIFO


__all__ = ["UnsupportedOperation"]

#
# Internals
#

_WINERROR_NOT_READY = 21  # drive exists but is not accessible
_WINERROR_INVALID_NAME = 123  # fix for bpo-35306
_WINERROR_CANT_RESOLVE_FILENAME = 1921  # broken symlink pointing to itself

# EBADF - guard against macOS `stat` throwing EBADF
_IGNORED_ERRNOS = (ENOENT, ENOTDIR, EBADF, ELOOP)

_IGNORED_WINERRORS = (
    _WINERROR_NOT_READY,
    _WINERROR_INVALID_NAME,
    _WINERROR_CANT_RESOLVE_FILENAME)

def _ignore_error(exception):
    return (getattr(exception, 'errno', None) in _IGNORED_ERRNOS or
            getattr(exception, 'winerror', None) in _IGNORED_WINERRORS)


@functools.cache
def _is_case_sensitive(parser):
    return parser.normcase('Aa') == 'Aa'


class UnsupportedOperation(NotImplementedError):
    """An exception that is raised when an unsupported operation is called on
    a path object.
    """
    pass


class ParserBase:
    """Base class for path parsers, which do low-level path manipulation.

    Path parsers provide a subset of the os.path API, specifically those
    functions needed to provide PurePathBase functionality. Each PurePathBase
    subclass references its path parser via a 'parser' class attribute.

    Every method in this base class raises an UnsupportedOperation exception.
    """

    @classmethod
    def _unsupported_msg(cls, attribute):
        return f"{cls.__name__}.{attribute} is unsupported"

    @property
    def sep(self):
        """The character used to separate path components."""
        raise UnsupportedOperation(self._unsupported_msg('sep'))

    def join(self, path, *paths):
        """Join path segments."""
        raise UnsupportedOperation(self._unsupported_msg('join()'))

    def split(self, path):
        """Split the path into a pair (head, tail), where *head* is everything
        before the final path separator, and *tail* is everything after.
        Either part may be empty.
        """
        raise UnsupportedOperation(self._unsupported_msg('split()'))

    def splitdrive(self, path):
        """Split the path into a 2-item tuple (drive, tail), where *drive* is
        a device name or mount point, and *tail* is everything after the
        drive. Either part may be empty."""
        raise UnsupportedOperation(self._unsupported_msg('splitdrive()'))

    def normcase(self, path):
        """Normalize the case of the path."""
        raise UnsupportedOperation(self._unsupported_msg('normcase()'))

    def isabs(self, path):
        """Returns whether the path is absolute, i.e. unaffected by the
        current directory or drive."""
        raise UnsupportedOperation(self._unsupported_msg('isabs()'))


class PurePathBase:
    """Base class for pure path objects.

    This class *does not* provide several magic methods that are defined in
    its subclass PurePath. They are: __fspath__, __bytes__, __reduce__,
    __hash__, __eq__, __lt__, __le__, __gt__, __ge__. Its initializer and path
    joining methods accept only strings, not os.PathLike objects more broadly.
    """

    __slots__ = (
        # The `_raw_path` slot store a joined string path. This is set in the
        # `__init__()` method.
        '_raw_path',

        # The '_resolving' slot stores a boolean indicating whether the path
        # is being processed by `PathBase.resolve()`. This prevents duplicate
        # work from occurring when `resolve()` calls `stat()` or `readlink()`.
        '_resolving',
    )
    parser = ParserBase()
    _globber = _Globber

    def __init__(self, path, *paths):
        self._raw_path = self.parser.join(path, *paths) if paths else path
        if not isinstance(self._raw_path, str):
            raise TypeError(
                f"path should be a str, not {type(self._raw_path).__name__!r}")
        self._resolving = False

    def with_segments(self, *pathsegments):
        """Construct a new path object from any number of path-like objects.
        Subclasses may override this method to customize how new path objects
        are created from methods like `iterdir()`.
        """
        return type(self)(*pathsegments)

    def __str__(self):
        """Return the string representation of the path, suitable for
        passing to system calls."""
        return self._raw_path

    def as_posix(self):
        """Return the string representation of the path with forward (/)
        slashes."""
        return str(self).replace(self.parser.sep, '/')

    @property
    def drive(self):
        """The drive prefix (letter or UNC path), if any."""
        return self.parser.splitdrive(self.anchor)[0]

    @property
    def root(self):
        """The root of the path, if any."""
        return self.parser.splitdrive(self.anchor)[1]

    @property
    def anchor(self):
        """The concatenation of the drive and root, or ''."""
        return self._stack[0]

    @property
    def name(self):
        """The final path component, if any."""
        return self.parser.split(self._raw_path)[1]

    @property
    def suffix(self):
        """
        The final component's last suffix, if any.

        This includes the leading period. For example: '.txt'
        """
        name = self.name
        i = name.rfind('.')
        if 0 < i < len(name) - 1:
            return name[i:]
        else:
            return ''

    @property
    def suffixes(self):
        """
        A list of the final component's suffixes, if any.

        These include the leading periods. For example: ['.tar', '.gz']
        """
        name = self.name
        if name.endswith('.'):
            return []
        name = name.lstrip('.')
        return ['.' + suffix for suffix in name.split('.')[1:]]

    @property
    def stem(self):
        """The final path component, minus its last suffix."""
        name = self.name
        i = name.rfind('.')
        if 0 < i < len(name) - 1:
            return name[:i]
        else:
            return name

    def with_name(self, name):
        """Return a new path with the file name changed."""
        split = self.parser.split
        if split(name)[0]:
            raise ValueError(f"Invalid name {name!r}")
        return self.with_segments(split(self._raw_path)[0], name)

    def with_stem(self, stem):
        """Return a new path with the stem changed."""
        suffix = self.suffix
        if not suffix:
            return self.with_name(stem)
        elif not stem:
            # If the suffix is non-empty, we can't make the stem empty.
            raise ValueError(f"{self!r} has a non-empty suffix")
        else:
            return self.with_name(stem + suffix)

    def with_suffix(self, suffix):
        """Return a new path with the file suffix changed.  If the path
        has no suffix, add given suffix.  If the given suffix is an empty
        string, remove the suffix from the path.
        """
        stem = self.stem
        if not stem:
            # If the stem is empty, we can't make the suffix non-empty.
            raise ValueError(f"{self!r} has an empty name")
        elif suffix and not (suffix.startswith('.') and len(suffix) > 1):
            raise ValueError(f"Invalid suffix {suffix!r}")
        else:
            return self.with_name(stem + suffix)

    def relative_to(self, other, *, walk_up=False):
        """Return the relative path to another path identified by the passed
        arguments.  If the operation is not possible (because this is not
        related to the other path), raise ValueError.

        The *walk_up* parameter controls whether `..` may be used to resolve
        the path.
        """
        if not isinstance(other, PurePathBase):
            other = self.with_segments(other)
        anchor0, parts0 = self._stack
        anchor1, parts1 = other._stack
        if anchor0 != anchor1:
            raise ValueError(f"{self._raw_path!r} and {other._raw_path!r} have different anchors")
        while parts0 and parts1 and parts0[-1] == parts1[-1]:
            parts0.pop()
            parts1.pop()
        for part in parts1:
            if not part or part == '.':
                pass
            elif not walk_up:
                raise ValueError(f"{self._raw_path!r} is not in the subpath of {other._raw_path!r}")
            elif part == '..':
                raise ValueError(f"'..' segment in {other._raw_path!r} cannot be walked")
            else:
                parts0.append('..')
        return self.with_segments('', *reversed(parts0))

    def is_relative_to(self, other):
        """Return True if the path is relative to another path or False.
        """
        if not isinstance(other, PurePathBase):
            other = self.with_segments(other)
        anchor0, parts0 = self._stack
        anchor1, parts1 = other._stack
        if anchor0 != anchor1:
            return False
        while parts0 and parts1 and parts0[-1] == parts1[-1]:
            parts0.pop()
            parts1.pop()
        for part in parts1:
            if part and part != '.':
                return False
        return True

    @property
    def parts(self):
        """An object providing sequence-like access to the
        components in the filesystem path."""
        anchor, parts = self._stack
        if anchor:
            parts.append(anchor)
        return tuple(reversed(parts))

    def joinpath(self, *pathsegments):
        """Combine this path with one or several arguments, and return a
        new path representing either a subpath (if all arguments are relative
        paths) or a totally different path (if one of the arguments is
        anchored).
        """
        return self.with_segments(self._raw_path, *pathsegments)

    def __truediv__(self, key):
        try:
            return self.with_segments(self._raw_path, key)
        except TypeError:
            return NotImplemented

    def __rtruediv__(self, key):
        try:
            return self.with_segments(key, self._raw_path)
        except TypeError:
            return NotImplemented

    @property
    def _stack(self):
        """
        Split the path into a 2-tuple (anchor, parts), where *anchor* is the
        uppermost parent of the path (equivalent to path.parents[-1]), and
        *parts* is a reversed list of parts following the anchor.
        """
        split = self.parser.split
        path = self._raw_path
        parent, name = split(path)
        names = []
        while path != parent:
            names.append(name)
            path = parent
            parent, name = split(path)
        return path, names

    @property
    def parent(self):
        """The logical parent of the path."""
        path = self._raw_path
        parent = self.parser.split(path)[0]
        if path != parent:
            parent = self.with_segments(parent)
            parent._resolving = self._resolving
            return parent
        return self

    @property
    def parents(self):
        """A sequence of this path's logical parents."""
        split = self.parser.split
        path = self._raw_path
        parent = split(path)[0]
        parents = []
        while path != parent:
            parents.append(self.with_segments(parent))
            path = parent
            parent = split(path)[0]
        return tuple(parents)

    def is_absolute(self):
        """True if the path is absolute (has both a root and, if applicable,
        a drive)."""
        return self.parser.isabs(self._raw_path)

    @property
    def _pattern_str(self):
        """The path expressed as a string, for use in pattern-matching."""
        return str(self)

    def match(self, path_pattern, *, case_sensitive=None):
        """
        Return True if this path matches the given pattern. If the pattern is
        relative, matching is done from the right; otherwise, the entire path
        is matched. The recursive wildcard '**' is *not* supported by this
        method.
        """
        if not isinstance(path_pattern, PurePathBase):
            path_pattern = self.with_segments(path_pattern)
        if case_sensitive is None:
            case_sensitive = _is_case_sensitive(self.parser)
        sep = path_pattern.parser.sep
        path_parts = self.parts[::-1]
        pattern_parts = path_pattern.parts[::-1]
        if not pattern_parts:
            raise ValueError("empty pattern")
        if len(path_parts) < len(pattern_parts):
            return False
        if len(path_parts) > len(pattern_parts) and path_pattern.anchor:
            return False
        globber = self._globber(sep, case_sensitive)
        for path_part, pattern_part in zip(path_parts, pattern_parts):
            match = globber.compile(pattern_part)
            if match(path_part) is None:
                return False
        return True

    def full_match(self, pattern, *, case_sensitive=None):
        """
        Return True if this path matches the given glob-style pattern. The
        pattern is matched against the entire path.
        """
        if not isinstance(pattern, PurePathBase):
            pattern = self.with_segments(pattern)
        if case_sensitive is None:
            case_sensitive = _is_case_sensitive(self.parser)
        globber = self._globber(pattern.parser.sep, case_sensitive, recursive=True)
        match = globber.compile(pattern._pattern_str)
        return match(self._pattern_str) is not None



class PathBase(PurePathBase):
    """Base class for concrete path objects.

    This class provides dummy implementations for many methods that derived
    classes can override selectively; the default implementations raise
    UnsupportedOperation. The most basic methods, such as stat() and open(),
    directly raise UnsupportedOperation; these basic methods are called by
    other methods such as is_dir() and read_text().

    The Path class derives this class to implement local filesystem paths.
    Users may derive their own classes to implement virtual filesystem paths,
    such as paths in archive files or on remote storage systems.
    """
    __slots__ = ()

    # Maximum number of symlinks to follow in resolve()
    _max_symlinks = 40

    @classmethod
    def _unsupported_msg(cls, attribute):
        return f"{cls.__name__}.{attribute} is unsupported"

    def stat(self, *, follow_symlinks=True):
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        raise UnsupportedOperation(self._unsupported_msg('stat()'))

    def lstat(self):
        """
        Like stat(), except if the path points to a symlink, the symlink's
        status information is returned, rather than its target's.
        """
        return self.stat(follow_symlinks=False)


    # Convenience functions for querying the stat results

    def exists(self, *, follow_symlinks=True):
        """
        Whether this path exists.

        This method normally follows symlinks; to check whether a symlink exists,
        add the argument follow_symlinks=False.
        """
        try:
            self.stat(follow_symlinks=follow_symlinks)
        except OSError as e:
            if not _ignore_error(e):
                raise
            return False
        except ValueError:
            # Non-encodable path
            return False
        return True

    def is_dir(self, *, follow_symlinks=True):
        """
        Whether this path is a directory.
        """
        try:
            return S_ISDIR(self.stat(follow_symlinks=follow_symlinks).st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_file(self, *, follow_symlinks=True):
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        try:
            return S_ISREG(self.stat(follow_symlinks=follow_symlinks).st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_mount(self):
        """
        Check if this path is a mount point
        """
        # Need to exist and be a dir
        if not self.exists() or not self.is_dir():
            return False

        try:
            parent_dev = self.parent.stat().st_dev
        except OSError:
            return False

        dev = self.stat().st_dev
        if dev != parent_dev:
            return True
        ino = self.stat().st_ino
        parent_ino = self.parent.stat().st_ino
        return ino == parent_ino

    def is_symlink(self):
        """
        Whether this path is a symbolic link.
        """
        try:
            return S_ISLNK(self.lstat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_junction(self):
        """
        Whether this path is a junction.
        """
        # Junctions are a Windows-only feature, not present in POSIX nor the
        # majority of virtual filesystems. There is no cross-platform idiom
        # to check for junctions (using stat().st_mode).
        return False

    def is_block_device(self):
        """
        Whether this path is a block device.
        """
        try:
            return S_ISBLK(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_char_device(self):
        """
        Whether this path is a character device.
        """
        try:
            return S_ISCHR(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_fifo(self):
        """
        Whether this path is a FIFO.
        """
        try:
            return S_ISFIFO(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def is_socket(self):
        """
        Whether this path is a socket.
        """
        try:
            return S_ISSOCK(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e):
                raise
            # Path doesn't exist or is a broken symlink
            # (see http://web.archive.org/web/20200623061726/https://bitbucket.org/pitrou/pathlib/issues/12/ )
            return False
        except ValueError:
            # Non-encodable path
            return False

    def samefile(self, other_path):
        """Return whether other_path is the same or not as this file
        (as returned by os.path.samefile()).
        """
        st = self.stat()
        try:
            other_st = other_path.stat()
        except AttributeError:
            other_st = self.with_segments(other_path).stat()
        return (st.st_ino == other_st.st_ino and
                st.st_dev == other_st.st_dev)

    def open(self, mode='r', buffering=-1, encoding=None,
             errors=None, newline=None):
        """
        Open the file pointed to by this path and return a file object, as
        the built-in open() function does.
        """
        raise UnsupportedOperation(self._unsupported_msg('open()'))

    def read_bytes(self):
        """
        Open the file in bytes mode, read it, and close the file.
        """
        with self.open(mode='rb') as f:
            return f.read()

    def read_text(self, encoding=None, errors=None, newline=None):
        """
        Open the file in text mode, read it, and close the file.
        """
        with self.open(mode='r', encoding=encoding, errors=errors, newline=newline) as f:
            return f.read()

    def write_bytes(self, data):
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        with self.open(mode='wb') as f:
            return f.write(view)

    def write_text(self, data, encoding=None, errors=None, newline=None):
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str):
            raise TypeError('data must be str, not %s' %
                            data.__class__.__name__)
        with self.open(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
            return f.write(data)

    def iterdir(self):
        """Yield path objects of the directory contents.

        The children are yielded in arbitrary order, and the
        special entries '.' and '..' are not included.
        """
        raise UnsupportedOperation(self._unsupported_msg('iterdir()'))

    def _glob_selector(self, parts, case_sensitive, recurse_symlinks):
        if case_sensitive is None:
            case_sensitive = _is_case_sensitive(self.parser)
            case_pedantic = False
        else:
            # The user has expressed a case sensitivity choice, but we don't
            # know the case sensitivity of the underlying filesystem, so we
            # must use scandir() for everything, including non-wildcard parts.
            case_pedantic = True
        recursive = True if recurse_symlinks else _no_recurse_symlinks
        globber = self._globber(self.parser.sep, case_sensitive, case_pedantic, recursive)
        return globber.selector(parts)

    def glob(self, pattern, *, case_sensitive=None, recurse_symlinks=True):
        """Iterate over this subtree and yield all existing files (of any
        kind, including directories) matching the given relative pattern.
        """
        if not isinstance(pattern, PurePathBase):
            pattern = self.with_segments(pattern)
        anchor, parts = pattern._stack
        if anchor:
            raise NotImplementedError("Non-relative patterns are unsupported")
        select = self._glob_selector(parts, case_sensitive, recurse_symlinks)
        return select(self)

    def rglob(self, pattern, *, case_sensitive=None, recurse_symlinks=True):
        """Recursively yield all existing files (of any kind, including
        directories) matching the given relative pattern, anywhere in
        this subtree.
        """
        if not isinstance(pattern, PurePathBase):
            pattern = self.with_segments(pattern)
        pattern = '**' / pattern
        return self.glob(pattern, case_sensitive=case_sensitive, recurse_symlinks=recurse_symlinks)

    def walk(self, top_down=True, on_error=None, follow_symlinks=False):
        """Walk the directory tree from this directory, similar to os.walk()."""
        paths = [self]
        while paths:
            path = paths.pop()
            if isinstance(path, tuple):
                yield path
                continue
            dirnames = []
            filenames = []
            if not top_down:
                paths.append((path, dirnames, filenames))
            try:
                for child in path.iterdir():
                    try:
                        if child.is_dir(follow_symlinks=follow_symlinks):
                            if not top_down:
                                paths.append(child)
                            dirnames.append(child.name)
                        else:
                            filenames.append(child.name)
                    except OSError:
                        filenames.append(child.name)
            except OSError as error:
                if on_error is not None:
                    on_error(error)
                if not top_down:
                    while not isinstance(paths.pop(), tuple):
                        pass
                continue
            if top_down:
                yield path, dirnames, filenames
                paths += [path.joinpath(d) for d in reversed(dirnames)]

    def absolute(self):
        """Return an absolute version of this path
        No normalization or symlink resolution is performed.

        Use resolve() to resolve symlinks and remove '..' segments.
        """
        raise UnsupportedOperation(self._unsupported_msg('absolute()'))

    @classmethod
    def cwd(cls):
        """Return a new path pointing to the current working directory."""
        # We call 'absolute()' rather than using 'os.getcwd()' directly to
        # enable users to replace the implementation of 'absolute()' in a
        # subclass and benefit from the new behaviour here. This works because
        # os.path.abspath('.') == os.getcwd().
        return cls('').absolute()

    def expanduser(self):
        """ Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        raise UnsupportedOperation(self._unsupported_msg('expanduser()'))

    @classmethod
    def home(cls):
        """Return a new path pointing to expanduser('~').
        """
        return cls("~").expanduser()

    def readlink(self):
        """
        Return the path to which the symbolic link points.
        """
        raise UnsupportedOperation(self._unsupported_msg('readlink()'))
    readlink._supported = False

    def resolve(self, strict=False):
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it.
        """
        if self._resolving:
            return self
        path_root, parts = self._stack
        path = self.with_segments(path_root)
        try:
            path = path.absolute()
        except UnsupportedOperation:
            path_tail = []
        else:
            path_root, path_tail = path._stack
            path_tail.reverse()

        # If the user has *not* overridden the `readlink()` method, then symlinks are unsupported
        # and (in non-strict mode) we can improve performance by not calling `stat()`.
        querying = strict or getattr(self.readlink, '_supported', True)
        link_count = 0
        while parts:
            part = parts.pop()
            if not part or part == '.':
                continue
            if part == '..':
                if not path_tail:
                    if path_root:
                        # Delete '..' segment immediately following root
                        continue
                elif path_tail[-1] != '..':
                    # Delete '..' segment and its predecessor
                    path_tail.pop()
                    continue
            path_tail.append(part)
            if querying and part != '..':
                path = self.with_segments(path_root + self.parser.sep.join(path_tail))
                path._resolving = True
                try:
                    st = path.stat(follow_symlinks=False)
                    if S_ISLNK(st.st_mode):
                        # Like Linux and macOS, raise OSError(errno.ELOOP) if too many symlinks are
                        # encountered during resolution.
                        link_count += 1
                        if link_count >= self._max_symlinks:
                            raise OSError(ELOOP, "Too many symbolic links in path", self._raw_path)
                        target_root, target_parts = path.readlink()._stack
                        # If the symlink target is absolute (like '/etc/hosts'), set the current
                        # path to its uppermost parent (like '/').
                        if target_root:
                            path_root = target_root
                            path_tail.clear()
                        else:
                            path_tail.pop()
                        # Add the symlink target's reversed tail parts (like ['hosts', 'etc']) to
                        # the stack of unresolved path parts.
                        parts.extend(target_parts)
                        continue
                    elif parts and not S_ISDIR(st.st_mode):
                        raise NotADirectoryError(ENOTDIR, "Not a directory", self._raw_path)
                except OSError:
                    if strict:
                        raise
                    else:
                        querying = False
        return self.with_segments(path_root + self.parser.sep.join(path_tail))

    def symlink_to(self, target, target_is_directory=False):
        """
        Make this path a symlink pointing to the target path.
        Note the order of arguments (link, target) is the reverse of os.symlink.
        """
        raise UnsupportedOperation(self._unsupported_msg('symlink_to()'))

    def hardlink_to(self, target):
        """
        Make this path a hard link pointing to the same file as *target*.

        Note the order of arguments (self, target) is the reverse of os.link's.
        """
        raise UnsupportedOperation(self._unsupported_msg('hardlink_to()'))

    def touch(self, mode=0o666, exist_ok=True):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        raise UnsupportedOperation(self._unsupported_msg('touch()'))

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        """
        Create a new directory at this given path.
        """
        raise UnsupportedOperation(self._unsupported_msg('mkdir()'))

    def rename(self, target):
        """
        Rename this path to the target path.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise UnsupportedOperation(self._unsupported_msg('rename()'))

    def replace(self, target):
        """
        Rename this path to the target path, overwriting if that path exists.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise UnsupportedOperation(self._unsupported_msg('replace()'))

    def chmod(self, mode, *, follow_symlinks=True):
        """
        Change the permissions of the path, like os.chmod().
        """
        raise UnsupportedOperation(self._unsupported_msg('chmod()'))

    def lchmod(self, mode):
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        self.chmod(mode, follow_symlinks=False)

    def unlink(self, missing_ok=False):
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        raise UnsupportedOperation(self._unsupported_msg('unlink()'))

    def rmdir(self):
        """
        Remove this directory.  The directory must be empty.
        """
        raise UnsupportedOperation(self._unsupported_msg('rmdir()'))

    def owner(self, *, follow_symlinks=True):
        """
        Return the login name of the file owner.
        """
        raise UnsupportedOperation(self._unsupported_msg('owner()'))

    def group(self, *, follow_symlinks=True):
        """
        Return the group name of the file gid.
        """
        raise UnsupportedOperation(self._unsupported_msg('group()'))

    @classmethod
    def from_uri(cls, uri):
        """Return a new path from the given 'file' URI."""
        raise UnsupportedOperation(cls._unsupported_msg('from_uri()'))

    def as_uri(self):
        """Return the path as a URI."""
        raise UnsupportedOperation(self._unsupported_msg('as_uri()'))



# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import shutil
import subprocess
import sys
import tempfile
import textwrap

import portage
from portage import os
from portage.const import (
    PORTAGE_PYM_PATH,
    USER_CONFIG_PATH,
)
from portage.process import find_binary
from portage.tests import TestCase, CommandStep, FunctionStep
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class BinpkgFetchtestCase(TestCase):
    def testLocalFilePkgSyncUpdate(self):
        """
        Check handling of local file:// sync-uri and unnecessary BUILD_ID
        increments (bug #921208).
        """
        debug = False

        ebuilds = {
            "dev-libs/A-1::local": {
                "EAPI": "7",
                "SLOT": "0",
            },
        }

        playground = ResolverPlayground(ebuilds=ebuilds, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        trees = playground.trees
        bindb = trees[eroot]["bintree"].dbapi
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)

        portage_python = portage._python_interpreter
        emerge_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "emerge"),
        )

        tmppkgdir = tempfile.TemporaryDirectory()
        tmppkgdir_suffix = os.path.join(tmppkgdir.name, "binpkg")

        test_commands = (
            # Create a trivial binpkg first.
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + ("--oneshot", "--verbose", "--buildpkg", "dev-libs/A"),
            ),
            # Copy to a new PKGDIR which we'll use as PORTAGE_BINHOST
            # then delete the old PKGDIR.
            FunctionStep(
                function=lambda _: shutil.copytree(
                    bindb.bintree.pkgdir, tmppkgdir_suffix
                )
                or True,
            ),
            FunctionStep(
                function=lambda _: os.unlink(
                    os.path.join(
                        bindb.bintree.pkgdir, "dev-libs", "A", "A-1-1.gpkg.tar"
                    )
                )
                or True,
            ),
            # This should succeed if we've correctly saved it as A-1-1.gpkg.tar, not
            # A-1-2.gpkg.tar, and then also try to unpack the right filename.
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + ("--oneshot", "--verbose", "--getbinpkgonly", "dev-libs/A"),
            ),
            # Check whether the downloaded binpkg in PKGDIR has the correct
            # filename (-1) or an unnecessarily-incremented one (-2).
            FunctionStep(
                function=lambda i: self.assertTrue(
                    os.path.exists(
                        os.path.join(
                            bindb.bintree.pkgdir, "dev-libs", "A", "A-1-1.gpkg.tar"
                        )
                    ),
                    f"step {i}",
                )
            ),
        )

        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")

        path = settings.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PATH": path,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
            "PORTAGE_INST_GID": str(os.getgid()),
            "PORTAGE_INST_UID": str(os.getuid()),
            "FEATURES": "-pkgdir-index-trusted",
        }

        dirs = [
            playground.distdir,
            fake_bin,
            portage_tmpdir,
            user_config_dir,
            var_cache_edb,
        ]

        true_symlinks = ["chown", "chgrp"]

        needed_binaries = {
            "true": (find_binary("true"), True),
        }

        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(needed_binaries["true"][0], os.path.join(fake_bin, x))

            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(test_commands):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                env["PORTAGE_BINHOST"] = f"file:///{tmppkgdir_suffix}"
                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )
        finally:
            playground.debug = False
            playground.cleanup()
            tmppkgdir.cleanup()

    def testFetchBinpkgWithPkgPretend(self):
        """
        Make sure error handling for pkg_pretend w/ --getbinpkg works
        correctly (fixed by a3f0843a3256ffe3f1248b903702d5b58c7ac892).
        """
        debug = False

        pkg_pretend = textwrap.dedent(
            """
        S="${WORKDIR}"

        pkg_pretend() {
            einfo "Hello world!"
        }
        """
        )

        ebuilds = {
            "dev-libs/A-1::local": {
                "EAPI": "7",
                "SLOT": "0",
                "MISC_CONTENT": pkg_pretend,
            },
            "dev-libs/B-1::local": {
                "EAPI": "7",
                "SLOT": "0",
                "MISC_CONTENT": pkg_pretend,
            },
        }

        playground = ResolverPlayground(ebuilds=ebuilds, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        trees = playground.trees
        bindb = trees[eroot]["bintree"].dbapi
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)

        portage_python = portage._python_interpreter
        emerge_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "emerge"),
        )

        tmppkgdir = tempfile.TemporaryDirectory()
        tmppkgdir_suffix = os.path.join(tmppkgdir.name, "binpkg")

        test_commands = (
            # Create two trivial binpkgs first.
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + ("--oneshot", "--verbose", "--buildpkg", "dev-libs/A", "dev-libs/B"),
            ),
            # Copy to a new PKGDIR which we'll use as PORTAGE_BINHOST
            # then delete the old PKGDIR.
            FunctionStep(
                function=lambda _: shutil.copytree(
                    bindb.bintree.pkgdir, tmppkgdir_suffix
                )
                or True,
            ),
            FunctionStep(
                function=lambda _: os.unlink(
                    os.path.join(
                        bindb.bintree.pkgdir, "dev-libs", "A", "A-1-1.gpkg.tar"
                    )
                )
            ),
            FunctionStep(
                function=lambda _: os.unlink(
                    os.path.join(
                        bindb.bintree.pkgdir, "dev-libs", "B", "B-1-1.gpkg.tar"
                    )
                )
            ),
            # Both dev-libs/A and dev-libs/B have a (passing) pkg_pretend to force
            # taking the path that was broken before.
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "--verbose",
                    "--jobs=2",
                    "--getbinpkgonly",
                    "dev-libs/A",
                    "dev-libs/B",
                ),
            ),
        )

        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")

        path = settings.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PATH": path,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
            "PORTAGE_INST_GID": str(os.getgid()),
            "PORTAGE_INST_UID": str(os.getuid()),
            "FEATURES": "-parallel-fetch -pkgdir-index-trusted",
        }

        dirs = [
            playground.distdir,
            fake_bin,
            portage_tmpdir,
            user_config_dir,
            var_cache_edb,
        ]

        true_symlinks = ["chown", "chgrp"]

        needed_binaries = {
            "true": (find_binary("true"), True),
        }

        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(needed_binaries["true"][0], os.path.join(fake_bin, x))

            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(test_commands):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                env["PORTAGE_BINHOST"] = f"file:///{tmppkgdir_suffix}"
                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )
        finally:
            playground.debug = False
            playground.cleanup()
            tmppkgdir.cleanup()



# Copyright 2016-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class RuntimeCycleMergeOrderTestCase(TestCase):
    def testRuntimeCycleMergeOrder(self):
        ebuilds = {
            "app-misc/plugins-consumer-1": {
                "EAPI": "6",
                "DEPEND": "app-misc/plugin-b:=",
                "RDEPEND": "app-misc/plugin-b:=",
            },
            "app-misc/plugin-b-1": {
                "EAPI": "6",
                "RDEPEND": "app-misc/runtime-cycle-b",
                "PDEPEND": "app-misc/plugins-consumer",
            },
            "app-misc/runtime-cycle-b-1": {
                "RDEPEND": "app-misc/plugin-b app-misc/branch-b",
            },
            "app-misc/branch-b-1": {
                "RDEPEND": "app-misc/leaf-b app-misc/branch-c",
            },
            "app-misc/leaf-b-1": {},
            "app-misc/branch-c-1": {
                "RDEPEND": "app-misc/runtime-cycle-c app-misc/runtime-c",
            },
            "app-misc/runtime-cycle-c-1": {
                "RDEPEND": "app-misc/branch-c",
            },
            "app-misc/runtime-c-1": {
                "RDEPEND": "app-misc/branch-d",
            },
            "app-misc/branch-d-1": {
                "RDEPEND": "app-misc/leaf-d app-misc/branch-e",
            },
            "app-misc/branch-e-1": {
                "RDEPEND": "app-misc/leaf-e",
            },
            "app-misc/leaf-d-1": {},
            "app-misc/leaf-e-1": {},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/plugin-b"],
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    ("app-misc/leaf-b-1", "app-misc/leaf-d-1", "app-misc/leaf-e-1"),
                    ("app-misc/branch-d-1", "app-misc/branch-e-1"),
                    "app-misc/runtime-c-1",
                    ("app-misc/runtime-cycle-c-1", "app-misc/branch-c-1"),
                    "app-misc/branch-b-1",
                    ("app-misc/runtime-cycle-b-1", "app-misc/plugin-b-1"),
                    "app-misc/plugins-consumer-1",
                ],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBuildtimeRuntimeCycleMergeOrder(self):
        installed = {
            "dev-util/cmake-3.26.5-r2": {
                "EAPI": "8",
                "KEYWORDS": "x86",
                "DEPEND": "net-misc/curl",
                "RDEPEND": "net-misc/curl",
            },
            "net-dns/c-ares-1.21.0": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
                "RDEPEND": "net-dns/c-ares",
            },
            "net-misc/curl-8.4.0": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
                "DEPEND": """
                    net-dns/c-ares
                    http2? ( net-libs/nghttp2:= )
                """,
                "RDEPEND": """
                    net-dns/c-ares
                    http2? ( net-libs/nghttp2:= )
                 """,
            },
            "net-dns/c-ares-1.21.0": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
            },
        }

        binpkgs = {
            "net-misc/curl-8.4.0": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
                "IUSE": "http2",
                "USE": "http2",
                "DEPEND": """
                    net-dns/c-ares
                    http2? ( net-libs/nghttp2:= )
                """,
                "RDEPEND": """
                    net-dns/c-ares
                    http2? ( net-libs/nghttp2:= )
                """,
            },
            "dev-util/cmake-3.26.5-r2": {
                "EAPI": "8",
                "KEYWORDS": "x86",
                "DEPEND": "net-misc/curl",
                "RDEPEND": "net-misc/curl",
            },
        }

        ebuilds = {
            "dev-util/cmake-3.26.5-r2": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
                "DEPEND": "net-misc/curl",
                "RDEPEND": "net-misc/curl",
            },
            "dev-util/cmake-3.27.8": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "~x86",
                "DEPEND": "net-misc/curl",
                "RDEPEND": "net-misc/curl",
            },
            "net-dns/c-ares-1.21.0": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
            },
            "net-libs/nghttp2-1.57.0": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
                "BDEPEND": "dev-util/cmake",
                "RDEPEND": "net-dns/c-ares",
            },
            "net-misc/curl-8.4.0": {
                "EAPI": "8",
                "SLOT": "0",
                "KEYWORDS": "x86",
                "IUSE": "http2",
                "DEPEND": """
                    net-dns/c-ares
                    http2? ( net-libs/nghttp2:= )
                """,
                "RDEPEND": """
                    net-dns/c-ares
                    http2? ( net-libs/nghttp2:= )
                """,
            },
        }

        world = ("dev-util/cmake",)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--verbose": True,
                    "--update": True,
                    "--deep": True,
                    "--newuse": True,
                    "--usepkg": True,
                },
                success=True,
                # It would also work to punt the dev-util/cmake upgrade
                # until the end, given it's already installed.
                mergelist=[
                    "dev-util/cmake-3.27.8",
                    "net-libs/nghttp2-1.57.0",
                    "[binary]net-misc/curl-8.4.0",
                ],
            ),
        )

        playground = ResolverPlayground(
            world=world,
            installed=installed,
            binpkgs=binpkgs,
            ebuilds=ebuilds,
            debug=False,
            user_config={
                "make.conf": (
                    'ACCEPT_KEYWORDS="~x86"',
                    'USE="http2"',
                ),
            },
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2015-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import bisect
import collections


class DbapiProvidesIndex:
    """
    The DbapiProvidesIndex class is used to wrap existing dbapi
    interfaces, index packages by the sonames that they provide, and
    implement the dbapi.match method for SonameAtom instances. Since
    this class acts as a wrapper, it can be used conditionally, so that
    soname indexing overhead is avoided when soname dependency
    resolution is disabled.

    Since it's possible for soname atom match results to consist of
    packages with multiple categories or names, it is essential that
    Package.__lt__ behave meaningfully when Package.cp is dissimilar,
    so that match results will be correctly ordered by version for each
    value of Package.cp.
    """

    _copy_attrs = (
        "aux_get",
        "aux_update",
        "categories",
        "cpv_all",
        "cpv_exists",
        "cp_all",
        "cp_list",
        "getfetchsizes",
        "settings",
        "_aux_cache_keys",
        "_clear_cache",
        "_cpv_sort_ascending",
        "_iuse_implicit_cnstr",
        "_pkg_str",
        "_pkg_str_aux_keys",
    )

    def __init__(self, db):
        self._db = db
        for k in self._copy_attrs:
            try:
                setattr(self, k, getattr(db, k))
            except AttributeError:
                pass
        self._provides_index = collections.defaultdict(list)

    def match(self, atom, use_cache=DeprecationWarning):
        if atom.soname:
            result = self._match_soname(atom)
        else:
            result = self._db.match(atom)
        return result

    def _match_soname(self, atom):
        result = self._provides_index.get(atom)
        if result is None:
            result = []
        else:
            result = [pkg.cpv for pkg in result]
        return result

    def _provides_inject(self, pkg):
        index = self._provides_index
        for atom in pkg.provides:
            # Use bisect.insort for ordered match results.
            bisect.insort(index[atom], pkg)


class PackageDbapiProvidesIndex(DbapiProvidesIndex):
    """
    This class extends DbapiProvidesIndex in order to make it suitable
    for wrapping a PackageVirtualDbapi instance.
    """

    _copy_attrs = DbapiProvidesIndex._copy_attrs + ("clear", "get", "_cpv_map")

    def clear(self):
        self._db.clear()
        self._provides_index.clear()

    def __bool__(self):
        return bool(self._db)

    def __iter__(self):
        return iter(self._db)

    def __contains__(self, item):
        return item in self._db

    def match_pkgs(self, atom):
        return [self._db._cpv_map[cpv] for cpv in self.match(atom)]

    def cpv_inject(self, pkg):
        self._db.cpv_inject(pkg)
        self._provides_inject(pkg)

    def cpv_remove(self, pkg):
        self._db.cpv_remove(pkg)
        index = self._provides_index
        for atom in pkg.provides:
            items = index[atom]
            try:
                items.remove(pkg)
            except ValueError:
                pass
            if not items:
                del index[atom]



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""Portability shim for xattr support

Exported API is the xattr object with get/get_all/set/remove/list operations.
We do not include the functions that Python 3.3+ provides in the os module as
the signature there is different compared to xattr.

See the standard xattr module for more documentation:
        https://pypi.python.org/pypi/pyxattr
"""

import contextlib
import os
import subprocess

from portage.exception import OperationNotSupported


class _XattrGetAll:
    """Implement get_all() using list()/get() if there is no easy bulk method"""

    @classmethod
    def get_all(cls, item, nofollow=False, namespace=None):
        return [
            (name, cls.get(item, name, nofollow=nofollow, namespace=namespace))
            for name in cls.list(item, nofollow=nofollow, namespace=namespace)
        ]


class _XattrSystemCommands(_XattrGetAll):
    """Implement things with getfattr/setfattr"""

    @staticmethod
    def _parse_output(output):
        for line in output.readlines():
            if line.startswith(b"#"):
                continue
            line = line.rstrip()
            if not line:
                continue
            # The lines will have the format:
            # 	user.hex=0x12345
            # 	user.base64=0sAQAAAgAgAAAAAAAAAAAAAAAAAAA=
            # 	user.string="value0"
            # But since we don't do interpretation on the value (we just
            # save & restore it), don't bother with decoding here.
            yield line.split(b"=", 1)

    @staticmethod
    def _call(*args, **kwargs):
        proc = subprocess.Popen(*args, **kwargs)
        if proc.stdin:
            proc.stdin.close()
        proc.wait()
        return proc

    @classmethod
    def get(cls, item, name, nofollow=False, namespace=None):
        if namespace:
            name = f"{namespace}.{name}"
        cmd = ["getfattr", "--absolute-names", "-n", name, item]
        if nofollow:
            cmd += ["-h"]
        proc = cls._call(cmd, stdout=subprocess.PIPE)

        value = None
        for _, value in cls._parse_output(proc.stdout):
            break

        proc.stdout.close()
        return value

    @classmethod
    def set(cls, item, name, value, _flags=0, namespace=None):
        if namespace:
            name = f"{namespace}.{name}"
        cmd = ["setfattr", "-n", name, "-v", value, item]
        cls._call(cmd)

    @classmethod
    def remove(cls, item, name, nofollow=False, namespace=None):
        if namespace:
            name = f"{namespace}.{name}"
        cmd = ["setfattr", "-x", name, item]
        if nofollow:
            cmd += ["-h"]
        cls._call(cmd)

    @classmethod
    def list(cls, item, nofollow=False, namespace=None, _names_only=True):
        cmd = ["getfattr", "-d", "--absolute-names", item]
        if nofollow:
            cmd += ["-h"]
        cmd += ["-m", (f"^{namespace}[.]") if namespace else "-"]
        proc = cls._call(cmd, stdout=subprocess.PIPE)

        ret = []
        if namespace:
            namespace = f"{namespace}."
        for name, value in cls._parse_output(proc.stdout):
            if namespace:
                if name.startswith(namespace):
                    name = name[len(namespace) :]
                else:
                    continue
            if _names_only:
                ret.append(name)
            else:
                ret.append((name, value))

        proc.stdout.close()
        return ret

    @classmethod
    def get_all(cls, item, nofollow=False, namespace=None):
        return cls.list(item, nofollow=nofollow, namespace=namespace, _names_only=False)


class _XattrStub(_XattrGetAll):
    """Fake object since system doesn't support xattrs"""

    # pylint: disable=unused-argument

    @staticmethod
    def _raise():
        e = OSError("stub")
        e.errno = OperationNotSupported.errno
        raise e

    @classmethod
    def get(cls, item, name, nofollow=False, namespace=None):
        cls._raise()

    @classmethod
    def set(cls, item, name, value, flags=0, namespace=None):
        cls._raise()

    @classmethod
    def remove(cls, item, name, nofollow=False, namespace=None):
        cls._raise()

    @classmethod
    def list(cls, item, nofollow=False, namespace=None):
        cls._raise()


if hasattr(os, "getxattr"):
    # Easy as pie -- active python supports it.
    class xattr(_XattrGetAll):
        """Python >=3.3 and GNU/Linux"""

        # pylint: disable=unused-argument

        @staticmethod
        def get(item, name, nofollow=False, namespace=None):
            return os.getxattr(item, name, follow_symlinks=not nofollow)

        @staticmethod
        def set(item, name, value, flags=0, namespace=None):
            return os.setxattr(item, name, value, flags=flags)

        @staticmethod
        def remove(item, name, nofollow=False, namespace=None):
            return os.removexattr(item, name, follow_symlinks=not nofollow)

        @staticmethod
        def list(item, nofollow=False, namespace=None):
            return os.listxattr(item, follow_symlinks=not nofollow)

else:
    try:
        # Maybe we have the xattr module.
        import xattr

    except ImportError:
        try:
            # Maybe we have the attr package.
            with open(os.devnull, "wb") as f:
                subprocess.call(["getfattr", "--version"], stdout=f)
                subprocess.call(["setfattr", "--version"], stdout=f)
            xattr = _XattrSystemCommands

        except OSError:
            # Stub it out completely.
            xattr = _XattrStub


# Add a knob so code can take evasive action as needed.
XATTRS_WORKS = xattr != _XattrStub


@contextlib.contextmanager
def preserve_xattrs(path, nofollow=False, namespace=None):
    """Context manager to save/restore extended attributes on |path|

    If you want to rewrite a file (possibly replacing it with a new one), but
    want to preserve the extended attributes, this will do the trick.

    # First read all the extended attributes.
    with save_xattrs('/some/file'):
            ... rewrite the file ...
    # Now the extended attributes are restored as needed.
    """
    kwargs = {
        "nofollow": nofollow,
    }
    if namespace:
        # Compiled xattr python module does not like it when namespace=None.
        kwargs["namespace"] = namespace

    old_attrs = dict(xattr.get_all(path, **kwargs))
    try:
        yield
    finally:
        new_attrs = dict(xattr.get_all(path, **kwargs))
        for name, value in new_attrs.items():
            if name not in old_attrs:
                # Clear out new ones.
                xattr.remove(path, name, **kwargs)
            elif new_attrs[name] != old_attrs[name]:
                # Update changed ones.
                xattr.set(path, name, value, **kwargs)

        for name, value in old_attrs.items():
            if name not in new_attrs:
                # Re-add missing ones.
                xattr.set(path, name, value, **kwargs)



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class DepcleanUnavailableSlotTestCase(TestCase):
    def testDepcleanUnavailableSlot(self):
        """
        Test bug #445506, where we want to remove the slot
        for which the ebuild is no longer available, even
        though its version is higher.
        """

        ebuilds = {
            "sys-kernel/gentoo-sources-3.0.53": {"SLOT": "3.0.53", "KEYWORDS": "x86"},
        }

        installed = {
            "sys-kernel/gentoo-sources-3.0.53": {"SLOT": "3.0.53", "KEYWORDS": "x86"},
            "sys-kernel/gentoo-sources-3.2.21": {"SLOT": "3.2.21", "KEYWORDS": "x86"},
        }

        world = ["sys-kernel/gentoo-sources"]

        test_cases = (
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["sys-kernel/gentoo-sources-3.2.21"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

        # Now make the newer version available and verify that
        # the lower version is depcleaned.
        ebuilds.update(
            {
                "sys-kernel/gentoo-sources-3.2.21": {
                    "SLOT": "3.2.21",
                    "KEYWORDS": "x86",
                },
            }
        )

        test_cases = (
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                cleanlist=["sys-kernel/gentoo-sources-3.0.53"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 1999-2012, 2016 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import logging

import portage
from portage import os
from portage.localization import _
from portage.output import bold, colorize, yellow
from portage.util import writemsg_level


def chk_updated_cfg_files(eroot, config_protect):
    target_root = eroot
    result = list(portage.util.find_updated_config_files(target_root, config_protect))

    for x in result:
        writemsg_level(
            f"\n {colorize('WARN', '* ' + _('IMPORTANT:'))} ",
            level=logging.INFO,
            noiselevel=-1,
        )
        if not x[1]:  # it's a protected file
            writemsg_level(
                f"config file '{x[0]}' needs updating.\n",
                level=logging.INFO,
                noiselevel=-1,
            )
        else:  # it's a protected dir
            if len(x[1]) == 1:
                head, tail = os.path.split(x[1][0])
                tail = tail[len("._cfg0000_") :]
                fpath = os.path.join(head, tail)
                writemsg_level(
                    f"config file '{fpath}' needs updating.\n",
                    level=logging.INFO,
                    noiselevel=-1,
                )
            else:
                writemsg_level(
                    f"{len(x[1])} config files in '{x[0]}' need updating.\n",
                    level=logging.INFO,
                    noiselevel=-1,
                )

    if result:
        print(
            " "
            + yellow("*")
            + " See the "
            + colorize("INFORM", _("CONFIGURATION FILES"))
            + " and "
            + colorize("INFORM", _("CONFIGURATION FILES UPDATE TOOLS"))
        )
        print(
            " "
            + yellow("*")
            + " sections of the "
            + bold("emerge")
            + " "
            + _("man page to learn how to update config files.")
        )



# Copyright 2022-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class UpdateIfInstalledTestCase(TestCase):
    def testUpdateIfInstalledEmerge(self):
        installed = {
            "dev-lang/ghc-4": {},
            "dev-libs/larryware-3": {},
            "dev-libs/larryware-ng-3": {},
            "virtual/libc-1": {},
        }

        ebuilds = installed.copy()
        ebuilds.update(
            {
                "app-misc/cowsay-10": {},
                "dev-lang/ghc-5": {},
                "dev-libs/larryware-4": {},
                "dev-libs/larryware-ng-4": {"RDEPEND": ">=net-libs/moo-1"},
                "net-libs/moo-1": {},
            }
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, debug=False
        )

        test_cases = (
            # We should only try to update ghc when passed ghc and
            # --update-if-installed. We don't want larryware to appear here,
            # despite it being eligible for an upgrade otherwise with --update.
            ResolverPlaygroundTestCase(
                ["dev-lang/ghc"],
                mergelist=["dev-lang/ghc-5"],
                options={
                    "--update-if-installed": True,
                },
                success=True,
            ),
            # Only try to upgrade ghc even if passed another candidate,
            # as there's no upgrade due for it. We don't want to
            # reinstall virtual/libc for the sake of it.
            ResolverPlaygroundTestCase(
                ["dev-lang/ghc", "virtual/libc"],
                mergelist=["dev-lang/ghc-5"],
                options={
                    "--update-if-installed": True,
                },
                success=True,
            ),
            # Try to upgrade a package with no new versions available.
            # This is just checking we still have --update semantics.
            ResolverPlaygroundTestCase(
                ["virtual/libc"],
                mergelist=[],
                options={
                    "--update-if-installed": True,
                },
                success=True,
            ),
            # If a new package is given, we want to do nothing.
            ResolverPlaygroundTestCase(
                ["app-misc/cowsay"],
                mergelist=[],
                options={
                    "--update-if-installed": True,
                },
                success=True,
            ),
            # If a new package (app-misc/cowsay) is given combined with
            # a package eligible for an upgrade (dev-libs/larryware),
            # upgrade just the latter.
            ResolverPlaygroundTestCase(
                ["app-misc/cowsay", "dev-libs/larryware"],
                mergelist=["dev-libs/larryware-4"],
                options={
                    "--update-if-installed": True,
                },
                success=True,
            ),
            # Make sure that we can still pull in upgrades as
            # dependencies (net-libs/moo) of the package we requested
            # (dev-libs/larryware-ng).
            ResolverPlaygroundTestCase(
                ["dev-libs/larryware-ng"],
                mergelist=["net-libs/moo-1", "dev-libs/larryware-ng-4"],
                options={
                    "--update-if-installed": True,
                },
                success=True,
            ),
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2015-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class BuildIdProfileFormatTestCase(TestCase):
    def testBuildIdProfileFormat(self):
        profile = {
            "eapi": "5",
            "packages": ("=app-misc/A-1-2::test_repo",),
            "package.mask": ("<app-misc/A-1::test_repo",),
            "package.keywords": ("=app-misc/A-1::test_repo x86",),
            "package.unmask": (">=app-misc/A-1::test_repo",),
            "package.use": ("=app-misc/A-1::test_repo foo",),
            "package.use.mask": ("=app-misc/A-1::test_repo -foo",),
            "package.use.stable.mask": ("=app-misc/A-1::test_repo -foo",),
            "package.use.force": ("=app-misc/A-1::test_repo foo",),
            "package.use.stable.force": ("=app-misc/A-1::test_repo foo",),
            "package.provided": ("sys-libs/zlib-1.2.8-r1",),
        }

        repo_configs = {
            "test_repo": {
                "layout.conf": (
                    "profile-formats = build-id profile-repo-deps profile-set",
                ),
            }
        }

        user_config = {
            "make.conf": ('FEATURES="binpkg-multi-instance"',),
        }

        ebuilds = {
            "app-misc/A-1": {
                "EAPI": "5",
                "RDEPEND": "sys-libs/zlib dev-libs/B[foo]",
                "DEPEND": "sys-libs/zlib dev-libs/B[foo]",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "IUSE": "foo",
            },
        }

        binpkgs = (
            (
                "app-misc/A-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "1",
                    "BUILD_TIME": "1",
                    "RDEPEND": "sys-libs/zlib dev-libs/B[foo]",
                    "DEPEND": "sys-libs/zlib dev-libs/B[foo]",
                },
            ),
            (
                "app-misc/A-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "2",
                    "BUILD_TIME": "2",
                    "RDEPEND": "sys-libs/zlib dev-libs/B[foo]",
                    "DEPEND": "sys-libs/zlib dev-libs/B[foo]",
                },
            ),
            (
                "app-misc/A-1",
                {
                    "EAPI": "5",
                    "BUILD_ID": "3",
                    "BUILD_TIME": "3",
                    "RDEPEND": "sys-libs/zlib dev-libs/B[foo]",
                    "DEPEND": "sys-libs/zlib dev-libs/B[foo]",
                },
            ),
            (
                "dev-libs/B-1",
                {
                    "EAPI": "5",
                    "IUSE": "foo",
                    "USE": "",
                    "BUILD_ID": "1",
                    "BUILD_TIME": "1",
                },
            ),
            (
                "dev-libs/B-1",
                {
                    "EAPI": "5",
                    "IUSE": "foo",
                    "USE": "foo",
                    "BUILD_ID": "2",
                    "BUILD_TIME": "2",
                },
            ),
            (
                "dev-libs/B-1",
                {
                    "EAPI": "5",
                    "IUSE": "foo",
                    "USE": "",
                    "BUILD_ID": "3",
                    "BUILD_TIME": "3",
                },
            ),
        )

        installed = {
            "app-misc/A-1": {
                "EAPI": "5",
                "BUILD_ID": "1",
                "BUILD_TIME": "1",
                "RDEPEND": "sys-libs/zlib",
                "DEPEND": "sys-libs/zlib",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "IUSE": "foo",
                "USE": "foo",
                "BUILD_ID": "2",
                "BUILD_TIME": "2",
            },
        }

        world = ()

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--emptytree": True, "--usepkgonly": True},
                success=True,
                mergelist=["[binary]dev-libs/B-1-2", "[binary]app-misc/A-1-2"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                _user_config = user_config.copy()
                _user_config["make.conf"] += (f'BINPKG_FORMAT="{binpkg_format}"',)
                playground = ResolverPlayground(
                    debug=False,
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    repo_configs=repo_configs,
                    profile=profile,
                    user_config=_user_config,
                    world=world,
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    # Disable debug so that cleanup works.
                    # playground.debug = False
                    playground.cleanup()



# Copyright 2010-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util.digraph import digraph

# ~ from portage.util import noiselimit
import portage.util


class DigraphTest(TestCase):
    def _assertBFSEqual(self, result, expected):
        result_stack = list(result)
        result_stack.reverse()
        expected_stack = list(reversed(expected))
        result_compared = []
        expected_compared = []
        while result_stack:
            if not expected_stack:
                result_compared.append(result_stack.pop())
                self.assertEqual(result_compared, expected_compared)
            expected_set = expected_stack.pop()
            if not isinstance(expected_set, list):
                expected_set = [expected_set]
            expected_set = set(expected_set)
            while expected_set:
                if not result_stack:
                    expected_compared.extend(expected_set)
                    self.assertEqual(result_compared, expected_compared)
                obj = result_stack.pop()
                try:
                    expected_set.remove(obj)
                except KeyError:
                    expected_compared.extend(expected_set)
                    result_compared.append(obj)
                    self.assertEqual(result_compared, expected_compared)
                else:
                    expected_compared.append(obj)
                    result_compared.append(obj)
        if expected_stack:
            expected_set = expected_stack.pop()
            if not isinstance(expected_set, list):
                expected_set = [expected_set]
            expected_compared.extend(expected_set)
            self.assertEqual(result_compared, expected_compared)

    def testBackwardCompatibility(self):
        g = digraph()
        f = g.copy()
        g.addnode("A", None)
        self.assertEqual("A" in g, True)
        self.assertEqual(bool(g), True)
        self.assertEqual(g.allnodes(), ["A"])
        self.assertEqual(g.allzeros(), ["A"])
        self.assertEqual(g.hasnode("A"), True)

    def testDigraphEmptyGraph(self):
        g = digraph()
        f = g.clone()
        for x in g, f:
            self.assertEqual(bool(x), False)
            self.assertEqual(x.contains("A"), False)
            self.assertEqual(x.firstzero(), None)
            self.assertRaises(KeyError, x.remove, "A")
            x.delnode("A")
            self.assertEqual(list(x), [])
            self.assertEqual(x.get("A"), None)
            self.assertEqual(x.get("A", "default"), "default")
            self.assertEqual(x.all_nodes(), [])
            self.assertEqual(x.leaf_nodes(), [])
            self.assertEqual(x.root_nodes(), [])
            self.assertRaises(KeyError, x.child_nodes, "A")
            self.assertRaises(KeyError, x.parent_nodes, "A")
            self.assertEqual(x.hasallzeros(), True)
            self.assertRaises(KeyError, list, x.bfs("A"))
            self.assertRaises(KeyError, x.shortest_path, "A", "B")
            self.assertRaises(KeyError, x.remove_edge, "A", "B")
            self.assertEqual(x.get_cycles(), [])
            x.difference_update("A")
            portage.util.noiselimit = -2
            x.debug_print()
            portage.util.noiselimit = 0

    def testDigraphCircle(self):
        g = digraph()
        g.add("A", "B", -1)
        g.add("B", "C", 0)
        g.add("C", "D", 1)
        g.add("D", "A", 2)

        f = g.clone()
        h = digraph()
        h.update(f)
        for x in g, f, h:
            self.assertEqual(bool(x), True)
            self.assertEqual(x.contains("A"), True)
            self.assertEqual(x.firstzero(), None)
            self.assertRaises(KeyError, x.remove, "Z")
            x.delnode("Z")
            self.assertEqual(list(x), ["A", "B", "C", "D"])
            self.assertEqual(x.get("A"), "A")
            self.assertEqual(x.get("A", "default"), "A")
            self.assertEqual(x.all_nodes(), ["A", "B", "C", "D"])
            self.assertEqual(x.leaf_nodes(), [])
            self.assertEqual(x.root_nodes(), [])
            self.assertEqual(x.child_nodes("A"), ["D"])
            self.assertEqual(x.child_nodes("A", ignore_priority=2), [])
            self.assertEqual(x.parent_nodes("A"), ["B"])
            self.assertEqual(x.parent_nodes("A", ignore_priority=-2), ["B"])
            self.assertEqual(x.parent_nodes("A", ignore_priority=-1), [])
            self.assertEqual(x.hasallzeros(), False)
            self._assertBFSEqual(
                x.bfs("A"), [(None, "A"), ("A", "D"), ("D", "C"), ("C", "B")]
            )
            self.assertEqual(x.shortest_path("A", "D"), ["A", "D"])
            self.assertEqual(x.shortest_path("D", "A"), ["D", "C", "B", "A"])
            self.assertEqual(x.shortest_path("A", "D", ignore_priority=2), None)
            self.assertEqual(
                x.shortest_path("D", "A", ignore_priority=-2), ["D", "C", "B", "A"]
            )
            cycles = {tuple(y) for y in x.get_cycles()}
            self.assertEqual(
                cycles,
                {
                    ("D", "C", "B", "A"),
                    ("C", "B", "A", "D"),
                    ("B", "A", "D", "C"),
                    ("A", "D", "C", "B"),
                },
            )
            x.remove_edge("A", "B")
            self.assertEqual(x.get_cycles(), [])
            x.difference_update(["D"])
            self.assertEqual(x.all_nodes(), ["A", "B", "C"])
            portage.util.noiselimit = -2
            x.debug_print()
            portage.util.noiselimit = 0

    def testDigraphTree(self):
        g = digraph()
        g.add("B", "A", -1)
        g.add("C", "A", 0)
        g.add("D", "C", 1)
        g.add("E", "C", 2)

        f = g.clone()
        for x in g, f:
            self.assertEqual(bool(x), True)
            self.assertEqual(x.contains("A"), True)
            self.assertEqual(x.has_edge("B", "A"), True)
            self.assertEqual(x.has_edge("A", "B"), False)
            self.assertEqual(x.firstzero(), "B")
            self.assertRaises(KeyError, x.remove, "Z")
            x.delnode("Z")
            self.assertEqual(set(x), {"A", "B", "C", "D", "E"})
            self.assertEqual(x.get("A"), "A")
            self.assertEqual(x.get("A", "default"), "A")
            self.assertEqual(set(x.all_nodes()), {"A", "B", "C", "D", "E"})
            self.assertEqual(set(x.leaf_nodes()), {"B", "D", "E"})
            self.assertEqual(set(x.leaf_nodes(ignore_priority=0)), {"A", "B", "D", "E"})
            self.assertEqual(x.root_nodes(), ["A"])
            self.assertEqual(set(x.root_nodes(ignore_priority=0)), {"A", "B", "C"})
            self.assertEqual(set(x.child_nodes("A")), {"B", "C"})
            self.assertEqual(x.child_nodes("A", ignore_priority=2), [])
            self.assertEqual(x.parent_nodes("B"), ["A"])
            self.assertEqual(x.parent_nodes("B", ignore_priority=-2), ["A"])
            self.assertEqual(x.parent_nodes("B", ignore_priority=-1), [])
            self.assertEqual(x.hasallzeros(), False)
            self._assertBFSEqual(
                x.bfs("A"),
                [(None, "A"), [("A", "C"), ("A", "B")], [("C", "E"), ("C", "D")]],
            )
            self.assertEqual(x.shortest_path("A", "D"), ["A", "C", "D"])
            self.assertEqual(x.shortest_path("D", "A"), None)
            self.assertEqual(x.shortest_path("A", "D", ignore_priority=2), None)
            cycles = {tuple(y) for y in x.get_cycles()}
            self.assertEqual(cycles, set())
            x.remove("D")
            self.assertEqual(set(x.all_nodes()), {"A", "B", "C", "E"})
            x.remove("C")
            self.assertEqual(set(x.all_nodes()), {"A", "B", "E"})
            portage.util.noiselimit = -2
            x.debug_print()
            portage.util.noiselimit = 0
            self.assertRaises(KeyError, x.remove_edge, "A", "E")

    def testDigraphCompleteGraph(self):
        g = digraph()
        g.add("A", "B", -1)
        g.add("B", "A", 1)
        g.add("A", "C", 1)
        g.add("C", "A", -1)
        g.add("C", "B", 1)
        g.add("B", "C", 1)

        f = g.clone()
        for x in g, f:
            self.assertEqual(bool(x), True)
            self.assertEqual(x.contains("A"), True)
            self.assertEqual(x.firstzero(), None)
            self.assertRaises(KeyError, x.remove, "Z")
            x.delnode("Z")
            self.assertEqual(list(x), ["A", "B", "C"])
            self.assertEqual(x.get("A"), "A")
            self.assertEqual(x.get("A", "default"), "A")
            self.assertEqual(x.all_nodes(), ["A", "B", "C"])
            self.assertEqual(x.leaf_nodes(), [])
            self.assertEqual(x.root_nodes(), [])
            self.assertEqual(set(x.child_nodes("A")), {"B", "C"})
            self.assertEqual(x.child_nodes("A", ignore_priority=0), ["B"])
            self.assertEqual(set(x.parent_nodes("A")), {"B", "C"})
            self.assertEqual(x.parent_nodes("A", ignore_priority=0), ["C"])
            self.assertEqual(x.parent_nodes("A", ignore_priority=1), [])
            self.assertEqual(x.hasallzeros(), False)
            self._assertBFSEqual(x.bfs("A"), [(None, "A"), [("A", "C"), ("A", "B")]])
            self.assertEqual(x.shortest_path("A", "C"), ["A", "C"])
            self.assertEqual(x.shortest_path("C", "A"), ["C", "A"])
            self.assertEqual(
                x.shortest_path("A", "C", ignore_priority=0), ["A", "B", "C"]
            )
            self.assertEqual(x.shortest_path("C", "A", ignore_priority=0), ["C", "A"])
            cycles = {frozenset(y) for y in x.get_cycles()}
            self.assertEqual(
                cycles,
                {
                    frozenset(["A", "B"]),
                    frozenset(["A", "C"]),
                    frozenset(["B", "C"]),
                },
            )
            x.remove_edge("A", "B")
            cycles = {frozenset(y) for y in x.get_cycles()}
            self.assertEqual(cycles, {frozenset(["A", "C"]), frozenset(["C", "B"])})
            x.difference_update(["C"])
            self.assertEqual(x.all_nodes(), ["A", "B"])
            portage.util.noiselimit = -2
            x.debug_print()
            portage.util.noiselimit = 0

    def testDigraphIgnorePriority(self):
        def always_true(dummy):
            return True

        def always_false(dummy):
            return False

        g = digraph()
        g.add("A", "B")

        self.assertEqual(g.parent_nodes("A"), ["B"])
        self.assertEqual(g.parent_nodes("A", ignore_priority=always_false), ["B"])
        self.assertEqual(g.parent_nodes("A", ignore_priority=always_true), [])

        self.assertEqual(g.child_nodes("B"), ["A"])
        self.assertEqual(g.child_nodes("B", ignore_priority=always_false), ["A"])
        self.assertEqual(g.child_nodes("B", ignore_priority=always_true), [])

        self.assertEqual(g.leaf_nodes(), ["A"])
        self.assertEqual(g.leaf_nodes(ignore_priority=always_false), ["A"])
        self.assertEqual(g.leaf_nodes(ignore_priority=always_true), ["A", "B"])

        self.assertEqual(g.root_nodes(), ["B"])
        self.assertEqual(g.root_nodes(ignore_priority=always_false), ["B"])
        self.assertEqual(g.root_nodes(ignore_priority=always_true), ["A", "B"])



# SPDX-License-Identifier: Apache-2.0
# Copyright 2012-2021 The Meson development team
# Copyright © 2021-2023 Intel Corporation

from __future__ import annotations

"""Posix specific implementations of mesonlib functionality."""

import fcntl
import typing as T

from .core import MesonException
from .platform import DirectoryLockBase, DirectoryLockAction

__all__ = ['DirectoryLock', 'DirectoryLockAction']

class DirectoryLock(DirectoryLockBase):

    def __enter__(self) -> None:
        try:
            self.lockfile = open(self.lockpath, 'w+', encoding='utf-8')
        except (FileNotFoundError, IsADirectoryError):
            # For FileNotFoundError, there is nothing to lock.
            # For IsADirectoryError, something is seriously wrong.
            raise
        except OSError:
            if self.action == DirectoryLockAction.IGNORE or self.optional:
                return

        try:
            flags = fcntl.LOCK_EX
            if self.action != DirectoryLockAction.WAIT:
                flags = flags | fcntl.LOCK_NB
            fcntl.flock(self.lockfile, flags)
        except BlockingIOError:
            self.lockfile.close()
            if self.action == DirectoryLockAction.IGNORE:
                return
            raise MesonException(self.err)
        except PermissionError:
            self.lockfile.close()
            raise MesonException(self.err)
        except OSError as e:
            self.lockfile.close()
            raise MesonException(f'Failed to lock directory {self.lockpath}: {e.strerror}')

    def __exit__(self, *args: T.Any) -> None:
        if self.lockfile is None or self.lockfile.closed:
            return
        fcntl.flock(self.lockfile, fcntl.LOCK_UN)
        self.lockfile.close()



# Copyright 2005-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

# parser option data
CHECK = {
    "short": "-c",
    "long": "--check",
    "help": "Check for problems (a default option for most modules)",
    "status": "Checking %s for problems",
    "action": "store_true",
    "func": "check",
}

FIX = {
    "short": "-f",
    "long": "--fix",
    "help": "Attempt to fix problems (a default option for most modules)",
    "status": "Attempting to fix %s",
    "action": "store_true",
    "func": "fix",
}

VERSION = {
    "long": "--version",
    "help": "show program's version number and exit",
    "action": "store_true",
}

# parser options
DEFAULT_OPTIONS = {"check": CHECK, "fix": FIX, "version": VERSION}



# Copyright 2012-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import shlex
import subprocess
import sys

import portage
from portage import os
from portage.const import BASH_BINARY, PORTAGE_PYM_PATH, USER_CONFIG_PATH
from portage.process import find_binary
from portage.tests import TestCase, CommandStep, FunctionStep
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class SlotAbiEmergeTestCase(TestCase):
    def testSlotAbiEmerge(self):
        debug = False

        ebuilds = {
            "dev-libs/glib-1.2.10": {"SLOT": "1"},
            "dev-libs/glib-2.30.2": {"EAPI": "5", "SLOT": "2/2.30"},
            "dev-libs/glib-2.32.3": {"EAPI": "5", "SLOT": "2/2.32"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2=",
                "RDEPEND": "dev-libs/glib:2=",
            },
        }
        installed = {
            "dev-libs/glib-1.2.10": {"EAPI": "5", "SLOT": "1"},
            "dev-libs/glib-2.30.2": {"EAPI": "5", "SLOT": "2/2.30"},
            "dev-libs/dbus-glib-0.98": {
                "EAPI": "5",
                "DEPEND": "dev-libs/glib:2/2.30=",
                "RDEPEND": "dev-libs/glib:2/2.30=",
            },
        }

        world = ["dev-libs/glib:1", "dev-libs/dbus-glib"]

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=debug
        )
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        trees = playground.trees
        portdb = trees[eroot]["porttree"].dbapi
        vardb = trees[eroot]["vartree"].dbapi
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)
        package_mask_path = os.path.join(user_config_dir, "package.mask")

        portage_python = portage._python_interpreter
        ebuild_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "ebuild"),
        )
        emerge_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "emerge"),
        )

        test_ebuild = portdb.findname("dev-libs/dbus-glib-0.98")
        self.assertFalse(test_ebuild is None)

        test_commands = (
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "dev-libs/glib",
                ),
            ),
            FunctionStep(
                function=lambda i: self.assertTrue(
                    "dev-libs/glib:2/2.32="
                    in vardb.aux_get("dev-libs/dbus-glib-0.98", ["RDEPEND"])[0],
                    f"step {i}",
                )
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=(BASH_BINARY,)
                + (
                    "-c",
                    "echo %s >> %s"
                    % tuple(
                        map(
                            shlex.quote,
                            (
                                ">=dev-libs/glib-2.32",
                                package_mask_path,
                            ),
                        )
                    ),
                ),
            ),
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "dev-libs/glib",
                ),
            ),
            FunctionStep(
                function=lambda i: self.assertTrue(
                    "dev-libs/glib:2/2.30="
                    in vardb.aux_get("dev-libs/dbus-glib-0.98", ["RDEPEND"])[0],
                    f"step {i}",
                )
            ),
        )

        distdir = playground.distdir
        pkgdir = playground.pkgdir
        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")
        profile_path = settings.profile_path

        path = settings.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PATH": path,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
            "PORTAGE_INST_GID": str(os.getgid()),
            "PORTAGE_INST_UID": str(os.getuid()),
        }

        if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
            env["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                "__PORTAGE_TEST_HARDLINK_LOCKS"
            ]

        dirs = [distdir, fake_bin, portage_tmpdir, user_config_dir, var_cache_edb]
        true_symlinks = ["chown", "chgrp"]
        true_binary = find_binary("true")
        self.assertEqual(true_binary is None, False, "true command not found")
        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(true_binary, os.path.join(fake_bin, x))
            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")
            # non-empty system set keeps --depclean quiet
            with open(os.path.join(profile_path, "packages"), "w") as f:
                f.write("*dev-libs/token-system-pkg")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(test_commands):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )
        finally:
            playground.cleanup()



# Copyright 2016-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import subprocess
import sys
import textwrap

import portage
from portage import os
from portage.const import PORTAGE_PYM_PATH, USER_CONFIG_PATH
from portage.process import find_binary
from portage.tests import TestCase, CommandStep, FunctionStep
from portage.util import ensure_dirs

from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class LibcDepInjectEmergeTestCase(TestCase):
    def testLibcDepInjection(self):
        """
        Test whether the implicit libc dependency injection (bug #913628)
        is correctly added for only ebuilds installing an ELF binary.

        Based on BlockerFileCollisionEmergeTestCase.
        """
        debug = False

        install_elf = textwrap.dedent(
            """
        S="${WORKDIR}"

        src_install() {
            insinto /usr/bin
            # We need an ELF binary for the injection to trigger, so
            # use ${BASH} given we know it must be around for running ebuilds.
            cp "${BASH}" "${ED}"/usr/bin/${PN} || die
        }
        """
        )

        ebuilds = {
            "sys-libs/glibc-2.38": {
                "EAPI": "8",
                "MISC_CONTENT": install_elf,
            },
            "virtual/libc-1": {
                "EAPI": "8",
                "RDEPEND": "sys-libs/glibc",
            },
            "dev-libs/A-1": {
                "EAPI": "8",
                "MISC_CONTENT": install_elf,
            },
            "dev-libs/B-1": {
                "EAPI": "8",
            },
            "dev-libs/C-1": {
                "EAPI": "8",
                "MISC_CONTENT": install_elf,
            },
            "dev-libs/D-1": {
                "EAPI": "8",
            },
            "dev-libs/E-1": {
                "EAPI": "8",
                "RDEPEND": ">=dev-libs/D-1",
                "MISC_CONTENT": install_elf,
            },
        }

        world = ("dev-libs/A",)

        playground = ResolverPlayground(ebuilds=ebuilds, world=world, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)

        portage_python = portage._python_interpreter
        emerge_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "emerge"),
        )

        test_commands = (
            # If we install a package with an ELF but no libc provider is
            # installed, make sure we don't inject anything (we don't want
            # to have some bare RDEPEND with literally "[]").
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "dev-libs/C",
                ),
            ),
            FunctionStep(
                function=lambda i: self.assertFalse(
                    portage.util.grablines(
                        os.path.join(
                            eprefix, "var", "db", "pkg", "dev-libs", "C-1", "RDEPEND"
                        )
                    ),
                    f"step {i}",
                ),
            ),
            # (We need sys-libs/glibc pulled in and virtual/libc installed)
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "virtual/libc",
                ),
            ),
            # A package NOT installing an ELF binary shouldn't have an injected libc dep
            # Let's check the virtual/libc one as we already have to merge it to pull in
            # sys-libs/glibc, but we'll do a better check after too.
            FunctionStep(
                function=lambda i: self.assertFalse(
                    ">=sys-libs/glibc-2.38\n"
                    in portage.util.grablines(
                        os.path.join(
                            eprefix, "var", "db", "pkg", "virtual", "libc-1", "RDEPEND"
                        )
                    ),
                    f"step {i}",
                ),
            ),
            # A package NOT installing an ELF binary shouldn't have an injected libc dep
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "dev-libs/B",
                ),
            ),
            FunctionStep(
                function=lambda i: self.assertFalse(
                    portage.util.grablines(
                        os.path.join(
                            eprefix, "var", "db", "pkg", "dev-libs", "B-1", "RDEPEND"
                        )
                    ),
                    f"step {i}",
                )
            ),
            # A package installing an ELF binary should have an injected libc dep
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "dev-libs/A",
                ),
            ),
            FunctionStep(
                function=lambda i: self.assertTrue(
                    os.path.exists(os.path.join(eroot, "usr/bin/A")), f"step {i}"
                )
            ),
            FunctionStep(
                lambda i: self.assertTrue(
                    ">=sys-libs/glibc-2.38\n"
                    in portage.util.grablines(
                        os.path.join(
                            eprefix, "var", "db", "pkg", "dev-libs", "A-1", "RDEPEND"
                        )
                    ),
                    f"step {i}",
                )
            ),
            # Install glibc again because earlier, no libc was installed, so the injection
            # wouldn't have fired even if the "are we libc?" check was broken.
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "sys-libs/glibc",
                ),
            ),
            # We don't want the libc (sys-libs/glibc is the provider here)
            # to have an injected dep on itself
            FunctionStep(
                function=lambda i: self.assertFalse(
                    ">=sys-libs/glibc-2.38\n"
                    in portage.util.grablines(
                        os.path.join(
                            eprefix,
                            "var",
                            "db",
                            "pkg",
                            "sys-libs",
                            "glibc-2.38",
                            "RDEPEND",
                        )
                    ),
                    f"step {i}",
                )
            ),
            # Make sure we append to, not clobber, RDEPEND
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "dev-libs/E",
                ),
            ),
            FunctionStep(
                function=lambda i: self.assertEqual(
                    [">=dev-libs/D-1 >=sys-libs/glibc-2.38\n"],
                    portage.util.grablines(
                        os.path.join(
                            eprefix, "var", "db", "pkg", "dev-libs", "E-1", "RDEPEND"
                        )
                    ),
                    f"step {i}",
                )
            ),
        )

        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")
        profile_path = settings.profile_path

        path = settings.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PATH": path,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
            "PORTAGE_INST_GID": str(os.getgid()),
            "PORTAGE_INST_UID": str(os.getuid()),
            "FEATURES": "-qa-unresolved-soname-deps -preserve-libs -merge-sync",
        }

        dirs = [
            playground.distdir,
            fake_bin,
            portage_tmpdir,
            user_config_dir,
            var_cache_edb,
        ]

        true_symlinks = ["chown", "chgrp"]

        # We don't want to make pax-utils a hard-requirement for tests,
        # so if it's not found, skip the test rather than FAIL it.
        needed_binaries = {
            "true": (find_binary("true"), True),
            "scanelf": (find_binary("scanelf"), False),
            "find": (find_binary("find"), True),
        }

        for name, (path, mandatory) in needed_binaries.items():
            found = path is not None

            if not found:
                if mandatory:
                    self.assertIsNotNone(path, f"command {name} not found")
                else:
                    self.skipTest(f"{name} not found")

        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(needed_binaries["true"][0], os.path.join(fake_bin, x))

            # We need scanelf, find for the ELF parts (creating NEEDED)
            os.symlink(needed_binaries["scanelf"][0], os.path.join(fake_bin, "scanelf"))
            os.symlink(needed_binaries["find"][0], os.path.join(fake_bin, "find"))

            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")
            with open(os.path.join(profile_path, "packages"), "w") as f:
                f.write("*virtual/libc")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(test_commands):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )

            # Check that dev-libs/A doesn't get re-emerged via --changed-deps
            # after injecting the libc dep. We want to suppress the injected
            # dep in the changed-deps comparisons.
            k = ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--changed-deps": True,
                    "--deep": True,
                    "--update": True,
                    "--verbose": True,
                },
                success=True,
                mergelist=[],
            )
            playground.run_TestCase(k)
            self.assertEqual(k.test_success, True, k.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def testBinpkgLibcDepInjection(self):
        """
        Test whether the implicit libc dependency injection (bug #913628)
        correctly forces an upgrade to a newer glibc before merging a binpkg
        built against it.

        Based on BlockerFileCollisionEmergeTestCase.
        """
        debug = False

        install_elf = textwrap.dedent(
            """
        S="${WORKDIR}"

        src_install() {
            insinto /usr/bin
            # We need an ELF binary for the injection to trigger, so
            # use ${BASH} given we know it must be around for running ebuilds.
            cp "${BASH}" "${ED}"/usr/bin/${PN} || die
        }
        """
        )

        ebuilds = {
            "sys-libs/glibc-2.37": {
                "EAPI": "8",
                "MISC_CONTENT": install_elf,
            },
            "sys-libs/glibc-2.38": {
                "EAPI": "8",
                "MISC_CONTENT": install_elf,
            },
            "virtual/libc-1": {
                "EAPI": "8",
                "RDEPEND": "sys-libs/glibc",
            },
            "dev-libs/A-1": {
                "EAPI": "8",
                "MISC_CONTENT": install_elf,
            },
            "dev-libs/B-1": {
                "EAPI": "8",
            },
            "dev-libs/C-1": {
                "EAPI": "8",
                "MISC_CONTENT": install_elf,
            },
        }

        playground = ResolverPlayground(ebuilds=ebuilds, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)

        portage_python = portage._python_interpreter
        emerge_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(str(self.bindir), "emerge"),
        )

        test_commands = (
            # (We need sys-libs/glibc pulled in and virtual/libc installed)
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "virtual/libc",
                ),
            ),
            # A package installing an ELF binary should have an injected libc dep
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "dev-libs/A",
                ),
            ),
            FunctionStep(
                function=lambda i: self.assertTrue(
                    os.path.exists(os.path.join(eroot, "usr/bin/A")), f"step {i}"
                )
            ),
            FunctionStep(
                function=lambda i: self.assertTrue(
                    ">=sys-libs/glibc-2.38\n"
                    in portage.util.grablines(
                        os.path.join(
                            eprefix, "var", "db", "pkg", "dev-libs", "A-1", "RDEPEND"
                        )
                    ),
                    f"step {i}",
                )
            ),
            # Downgrade glibc to a version (2.37) older than the version
            # that dev-libs/A's binpkg was built against (2.38). Below,
            # we check that it pulls in a newer glibc via a ResolverPlayground
            # testcase.
            CommandStep(
                returncode=os.EX_OK,
                command=emerge_cmd
                + (
                    "--oneshot",
                    "--nodeps",
                    "<sys-libs/glibc-2.38",
                ),
            ),
        )

        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")
        profile_path = settings.profile_path

        path = settings.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PATH": path,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
            "PORTAGE_INST_GID": str(os.getgid()),
            "PORTAGE_INST_UID": str(os.getuid()),
            "FEATURES": "buildpkg",
        }

        dirs = [
            playground.distdir,
            fake_bin,
            portage_tmpdir,
            user_config_dir,
            var_cache_edb,
        ]

        true_symlinks = ["chown", "chgrp"]

        # We don't want to make pax-utils a hard-requirement for tests,
        # so if it's not found, skip the test rather than FAIL it.
        needed_binaries = {
            "true": (find_binary("true"), True),
            "scanelf": (find_binary("scanelf"), False),
            "find": (find_binary("find"), True),
        }

        for name, (path, mandatory) in needed_binaries.items():
            found = path is not None

            if not found:
                if mandatory:
                    self.assertIsNotNone(path, f"command {name} not found")
                else:
                    self.skipTest(f"{name} not found")

        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(needed_binaries["true"][0], os.path.join(fake_bin, x))

            # We need scanelf, find for the ELF parts (creating NEEDED)
            os.symlink(needed_binaries["scanelf"][0], os.path.join(fake_bin, "scanelf"))
            os.symlink(needed_binaries["find"][0], os.path.join(fake_bin, "find"))

            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")
            with open(os.path.join(profile_path, "packages"), "w") as f:
                f.write("*virtual/libc")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(test_commands):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )

            # Now check that glibc gets upgraded to the right version
            # for the binpkg first after we downgraded it earlier, before
            # merging the dev-libs/A binpkg which needs 2.38.
            k = ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={
                    "--usepkgonly": True,
                    "--verbose": True,
                },
                success=True,
                mergelist=["[binary]sys-libs/glibc-2.38-1", "[binary]dev-libs/A-1-1"],
            )
            playground.run_TestCase(k)
            self.assertEqual(k.test_success, True, k.fail_msg)

        finally:
            playground.debug = False
            playground.cleanup()
